'use strict'
/* app/add.js
 * route for a client to add a peer
*/

const env = require(process.argv[2]||'../env.json')
const helper = require('../includes/helpers')
const fs = require('fs').promises
const wg = require('../includes/wireguard')
const https = require('https')
// Construct axios in IIFE for async/await
let axios; (async()=>{
	try {
			axios = require('axios').create({
		  httpsAgent: new https.Agent({
		  	ca: await fs.readFile(env.CA_CERT_FILE),
		  }),
		})
	} catch (err) { console.error(err) }
})()
const dns_key = `hmac-sha512:wgapi-${env.LOCAL_SERVER}:${env.DNS_KEY}`
let config_queue = []

module.exports = async (req, res) => {
	
	// Parse, sanitize, and validate hostname
	const new_hostname = req.query['name'].trim().toLowerCase()
	if (!new_hostname) {
		console.log(`New peer request from ${req.requester} didn't provide a hostname`)
		return res.sendStatus(400)
	} else if (!/^([\-\_a-z0-9]{1,20})$/.test(new_hostname)) {
		console.log(`New peer request from ${req.requester} provided an invalid hostname: ${new_hostname}`)
		return res.sendStatus(400)
	} else console.log(`New peer request from ${req.requester} for ${new_hostname}`)
	
	// Get user from IP
	let user; try { user = await helper.getUserFromIp(req.requester) }
	catch (err) {
		console.error(`Failed to get user from ${req.requester}`)
		return res.sendStatus(err)
	}	console.log(`${req.requester} must be ${user.name}`)
	
	// Check token
	if (req.query['token']!==helper.getToken(req.requester)) {
		console.log(`Invalid token from ${req.requester}: ${req.query['token']}`)
		return res.sendStatus(403)
	}
	
	// Check if new peer already exists
	if (user.peers.map((peer) => peer.name).includes(new_hostname)) {
		console.log(`Host already exists for ${domain}`)
		return res.sendStatus(409)
	}
	
	const domain = `${new_hostname}.${user.name}.${env.TLD}`
	
	// Find next available host part
	const used_ipv4_hosts = user.peers
		.map((host) => host.ipv4).map((found_ipv4) => 
			found_ipv4.toString().split('.')[3])
	const used_ipv6_hosts = user.peers
		.map((host) => host.ipv6).map((found_ipv6) => 
			found_ipv6.toString().split(':')[3])
	let host = 1
	while ([...used_ipv4_hosts,...used_ipv6_hosts].includes(host.toString())) host++
	if (host>9999) {
		console.error(`New host part for ${user.name} is higher than 9999: ${host}`)
		return res.sendStatus(507)
	}
	
	// Create IP Addresses and keys
	let ipv4_addr; if (host<254)
		ipv4_addr = `${env.IPV4_NET}.${user.subnet}.${host}`
	const ipv6_addr = `${env.IPV6_NET}:${user.subnet}:${host}`
	let keypair; try {
		keypair = await wg.generateKeypair()
	} catch (err) {
		console.error(`Unable to generateKeypair:\n${err}`)
		return res.sendStatus(409)
	}

	// Peer with each server
	let client_peers = []
	for (const server of env.SERVERS) {
		let psk; try {
			psk = await wg.generatePSK()
		} catch (err) {
			console.error(`Unable to generate PSK:\n${err}`)
			return res.sendStatus(409)
		}
		
		// Add server to user config as [Peer]
		const allowed_ipv4s = `${server.ipv4}/${(server.host===env.LOCAL_SERVER)?env.IPV4_CIDR:'32'}`
		const allowed_ipv6s = `${server.ipv6}/${(server.host===env.LOCAL_SERVER)?env.IPV6_CIDR:'128'}`
		client_peers.push(`
[Peer] # ${server.host}.${env.TLD}
PublicKey = ${server.pubkey}
PresharedKey = ${psk}
AllowedIPs = ${allowed_ipv4s}, ${allowed_ipv6s}
Endpoint = ${server.endpoint}
PersistentKeepAlive = 25`)

		// Add new user device to server config as [Peer]
		const allowed_ips = ipv4_addr
			? `${ipv4_addr}/32, ${ipv6_addr}/128`
			: `${ipv6_addr}/128`
		const server_config = `\n
[Peer] # ${domain}
PublicKey = ${keypair[0]}
PresharedKey = ${psk}
AllowedIPs = ${allowed_ips}`

		// Add server_config to local wg0.conf
		if (server.host===env.LOCAL_SERVER) {
			try { await fs.appendFile(env.WG_CONFIG_FILE, server_config) }
			catch (err) { return console.error(err) }
			
		// Remote server: send config
		} else {
			console.log(`Sending config to ${server.host}.gf4`)
			try {
				await axios.post(`${server.admin_endpoint}/add?secret=${server.secret}`, server_config, {
					headers: {'Content-Type': 'text/plain'},
				})
			} catch (err) {
				if (err.message==='Request failed with status code 403') {
					console.error(`Received 403 from ${server.admin_endpoint}/add`)
				} else {
					console.error(`Failed to add peer to ${server.host}:\n${server_config}`)
					if (err) console.error(err.message)
				}
			}
		}
	}

	// Update nameserver
	if (env.ENV==='prod') {
		try {
			await helper.nsUpdate(dns_key, env.DNS_MASTER, 
`update add ${domain}. ${env.DNS_TTL} A ${ipv4_addr}
update add ${domain}. ${env.DNS_TTL} AAAA ${ipv6_addr}
update add *.${domain}. ${env.DNS_TTL} CNAME ${domain}.`)
		}
		catch (err) {
			console.error(`Failed to add ns record.`)
			if (err) console.error(err)
		}
		console.log(`Updated nameserver to add ${domain}.`)
	}

	// Generate user config
	const listen_port = Math.floor(50000 + Math.random() * 10000)
	const config_address = ipv4_addr
		? `${ipv4_addr}/${env.IPV4_CIDR}, ${ipv6_addr}/${env.IPV6_CIDR}`
		: `${ipv6_addr}/${env.IPV6_CIDR}`
	const config = `[Interface] # ${domain}
PrivateKey = ${keypair[1]}
Address = ${config_address}
DNS = ${res.locals.DNS_SERVERS_STRING}
ListenPort = ${listen_port}
PostUp = resolvectl domain ${env.TLD} ${env.TLD}
${client_peers.join('\n')}`

	// Send config to user
	res.setHeader('content-type', 'text/plain')
	return res.send(config)

}