'use strict'
/* app/del.js
 * route to let clients delete a peer
*/

const fs = require('fs').promises
const env = require(process.argv[2]||'../env.json')
const helper = require('../includes/helpers')
const wg = require('../includes/wireguard')
const https = require('https')
// Construct axios in IIFE for async/await
let axios; (async()=>{
	try {
			axios = require('axios').create({
		  httpsAgent: new https.Agent({
		  	ca: await fs.readFile(env.CA_CERT_FILE),
		  }),
		})
	} catch (err) { console.log(err) }
})()
const dns_key = `hmac-sha512:wgapi-${env.LOCAL_SERVER}:${env.DNS_KEY}`

module.exports = async (req, res) => {
	const target  = req.query['name']
		|| req.query['psk']
		|| req.query['ip']
		|| req.query['pubkey']
		|| req.query['privkey']
	console.log(`Received request from ${req.requester} to delete ${target}`)

	// Get user from IP
	let user; try { user = await helper.getUserFromIp(req.requester) }
	catch (err) {
		console.error(`Failed to get user from ${req.requester}`)
		return res.sendStatus(err)
	}
	console.log(`${req.requester} must be ${user.name}`)
	
	// Check token
	if (req.query['token']!==helper.getToken(req.requester)) {
		console.log(`Invalid token from ${req.requester}: ${req.query['token']}`)
		return res.sendStatus(403)
	}

	// Load wg.conf
	let config_file
	try { config_file = await fs.readFile(env.WG_CONFIG_FILE) }
	catch (err) { console.error(err); return res.sendStatus(500) }

	// Search for peer
	let peer_pubkey
	const config = config_file.toString()
	const peer = config.split('\n\n')
		.filter( (paragraph) => {
			return paragraph.includes('[Peer]')
		}).filter( (peer) => {
			if (req.query['name']) {
				return peer.includes(`[Peer] # ${req.query['name']}.`)
			} else if (req.query['pubkey']) {
				peer_pubkey = req.query['pubkey']
				return peer.includes(`PublicKey = ${req.query['pubkey']}`)
			} else if (req.query['psk']) {
				return peer.includes(`PresharedKey = ${req.query['psk']}`)
			} else if (req.query['ip']) {
				return peer.split('\n').some( (line) => (
					line.includes('AllowedIPs') &&
					line.includes(` ${req.query['ip']}/`)
				) )
			} else if (req.query['privkey']) {
				// .filter() doesn't support async so use then/catch in this block
				wg.getPubkeyFromPrivkey(req.query['privkey'])
				.then((pubkey) => {
					peer_pubkey = pubkey
					return peer.includes(`PublicKey = ${pubkey}`)
				})
				.catch((err) => {
					console.error(`Failed to generate public key from private key during delete request\n`,err)
					return res.sendStatus(500)
				})
			} else {
				console.error(`${req.requester} sent delete request without specifying a peer`)
				return res.sendStatus(400)
			}
		})[0]
	if (peer===undefined) {
		console.log(`No peer found for delete request from ${req.requester}`)
		return res.sendStatus(404)
	}
	
	// Parse peer
	const peer_lines = peer.split('\n')
	const peer_name = peer_lines
		.filter( (line) => line.includes('[Peer] # ') )[0]
		.split(' # ')[1]
	const peer_ips = peer_lines
		.filter( (line) => line.includes('AllowedIPs = '))[0]
		.split(' = ')[1].split(', ')
	if (peer_pubkey===undefined) {
		peer_pubkey = peer_lines
			.filter( (line) => line.includes('PublicKey = ') )[0]
			.split(' = ')[1]
	}
	
	// Make sure requester can't delete self
	if (peer_ips.includes(req.requester)) {
		console.log(`Refused to let ${req.requester} delete self: ${peer_name}`)
		return res.sendStatus(409) // "Conflict"
	}
	
	// Delete from local wg config
	console.log(`Deleting ${peer_name}`); try {
		await fs.writeFile(env.WG_CONFIG_FILE, 
			config.replace(`\n\n${peer}`,'')
			.replace('\n\n\n','\n\n')
		)
	} catch (err) {
		console.error(`Failed to delete ${peer_name}:\n`,err)
		return res.sendStatus(500)
	}
	
	// Inform other servers
	for (const server of env.SERVERS) {
		if (server.host!==env.LOCAL_SERVER) {
			try {
				console.log(`Informing ${server.host} to delete ${peer_name}`)
				await axios.post(`${server.admin_endpoint}/del?secret=${server.secret}`, peer_pubkey, {
					headers: {'Content-Type': 'text/plain'},
				})
			} catch (err) {
				if (err) console.error(err)
				console.error(`Failed to inform ${server.host} to delete ${peer_name}!`)
				return res.sendStatus(500)
			}
		}
	}
		
	// Delete domains from nameserver
	if (env.ENV==='prod') {
		try {
			await helper.nsUpdate(dns_key, env.DNS_MASTER, 
`update delete ${peer_name}. A
update delete ${peer_name}. AAAA
update delete *.${peer_name}. CNAME`)
		}
		catch (err) {
			console.error(`Failed to delete ns record`)
			if (err) console.error(err)
			return res.sendStatus(500)
		}
		console.log(`Updated nameserver to delete ${peer_name}.`)
	}
	
	return res.sendStatus(200)

}