'use strict'
const crypto = require('crypto')
const env = require(process.argv[2]||'../env.json')
const fs = require('fs').promises
const spawn = require('child_process').spawn

let tokens = {}

module.exports = {
	
	getUserFromIp: (ip) => 
		new Promise( async (resolve, reject) => {

			// Get subnet (number)
			let subnet
			if (ip.includes(env.IPV4_NET)) subnet = ip.split('.').slice(-2,-1)[0]
			else if (ip.includes(env.IPV6_NET)) subnet = ip.split(':').slice(-2,-1)[0]
			else console.log(`Received request from ${ip}, which does not appear to be from the network.`)
	
			// Read wg.conf file for this user's other devices
			let userpeers; try {
				userpeers = (await fs.readFile(env.WG_CONFIG_FILE)).toString()
					.split('\n\n').filter( (paragraph) => {
						return paragraph.includes('[Peer]')
					}).filter( (peer) => {
						return peer.includes(`${env.IPV4_NET}.${subnet}`) || peer.includes(`${env.IPV6_NET}:${subnet}`)
					})
				} catch (err) { console.log(err); reject(500) }
				let found_usernames = []
				let found_hosts = []
				for (const userpeer of userpeers) {
					let userpeer_obj = {}
					for (const line of userpeer.split('\n')) {
						if (line.includes('[Peer] # ')) {
							const domain = line.split(' # ')[1].split('.')
							userpeer_obj.name = domain[0]
							found_usernames.push(domain[1])
						}
						else if (line.includes('AllowedIPs = ')) {
							const ips = line.split('=')[1].split(', ')
							userpeer_obj.ipv4 = ips.filter( (ip) => ip.includes(env.IPV4_NET) )[0].trim().split('/')[0]
							userpeer_obj.ipv6 = ips.filter( (ip) => ip.includes(env.IPV6_NET) )[0].trim().split('/')[0]
						}
					}
					found_hosts.push(userpeer_obj)
				}
				
				// Check that all IP addresses are in correct subnet or error out
				if (!found_hosts.every((host) => 
					host.ipv4.includes(`${env.IPV4_NET}.${subnet}.`) &&
					host.ipv6.includes(`${env.IPV6_NET}:${subnet}:`)
				)) {
					console.error(
						`Found unmatching IP address subnets for ${ip}: \
						${found_hosts.map( (host) => [host.ipv4,host.ipv6] )}`
					); reject(500)
			  // Check that the ip is "on the list"
			  // Should never get here because this ip can't access this IP!
			  } else if (found_usernames.length ===0) {
			  	console.error(`Received request from ${ip} not in wg.conf!`)
					reject(403)
				// Check that all usernames are the same correct or error out
				// https://stackoverflow.com/a/35568895
				} else if (!found_usernames.every( (v,i,r) => v === r[0] )) {
					console.error(`Found unmatching usernames for ${ip}: ${found_usernames.toString()}`)
					reject(500)
				} else {
					resolve({
						name: found_usernames[0],
						subnet: subnet,
						peers: found_hosts,
					})
				}
	}),
	
	getToken: (ip) => tokens[ip],
	setToken: (ip) => 
		new Promise ( async (resolve, reject) => {
			try {
				tokens[ip] = await crypto.randomBytes(40).toString('hex')
			} catch (err) { reject(err) }
			resolve(tokens[ip])
	}),
	
	nsUpdate: (key, server, payload) => 
		new Promise( (resolve, reject) => {
			try {
				
				let nsupdate = spawn('nsupdate', ['-y', key])

				// Collect output
				let errors = ''
				nsupdate.stdout.on('data', (data) => {
					console.log(`nsupdate stdout: ${data}`)
				})
				nsupdate.stderr.on('data', (data) => {
					console.error(`nsupdate stderr: ${data}`)
					errors += data
				})
				
				// Send data
				nsupdate.stdin.write(`server ${server}\n${payload}\nsend\nquit`)
				nsupdate.stdin.end()				
				
				// Handle exit
				nsupdate.on('error', (err) => {
					console.error(`nsupdate threw an error`)
					if (err) console.error(err)
					reject(err)
				})
				nsupdate.on('exit', (status) => {
					console.log(`nsupdate exited with status: ${status}`)
					if (status!==0) reject(errors)
					else resolve()
				})
			
			// Something went wrong with the spawn?
			} catch (err) {
				console.error('nsupdate failed to spawn or something')
				if (err) console.error(err)
				reject(err)
			}
			
	}),
	
}